%%% FourBarP2P.m
%%%
%%% Run buildModel.m first to build the OSIM model of the linkage.
%%%
%%% Run this file to optimize the trajectory from position to position.
%%%
%%% Author:     Tyler Morrison
%%% Ohio State University - Dept. of Mechanical and Aerospace Engineering
%%% email:      morrison.730@osu.edu 
%%% Website:    http://www.tyler-morrison.com
%%% October 2020

%% Clear Workspace
clc; close all; clearvars;

%% Import OpenSim Libraries into Matlab
import org.opensim.modeling.*

%% Import model.
modelName = 'FourBarTest.osim';
analysis_name = 'fourBarP2P';
dirResults = '';
model = Model(modelName);
model.initSystem();
dateLabel = datestr(datetime('now'),'yyyy-mm-dd-HH-MM-SS');

% Move the angle of link C from start to end position in fixed time.
CO_rz_0 = deg2rad(90);  % START Position
CO_rz_1 = deg2rad(45);  % END   Position

%% Build Study and Problem
study = MocoStudy();
study.setName(analysis_name);

problem = study.updProblem();

modelProcessor = ModelProcessor(model);
problem.setModelProcessor(modelProcessor);

%% Goal
% Objective: Minimize effort.
effortGoal = MocoControlGoal('effort', 10);
effortGoal.setExponent(2);
problem.addGoal(effortGoal);

%% Bounds
problem.setStateInfo('/jointCO/CO_rz/value', ...
    [deg2rad(-180), deg2rad(180)], ... % Overall UB and LB.
    CO_rz_0, ... % Exact start position.
    CO_rz_1);    % Exact end position.
% Start and end at rest.
problem.setStateInfo('/jointCO/CO_rz/speed', [-50, 50], 0, 0);
% One second to do so.
problem.setTimeBounds(0.0, 1.0);
% A guess at reasonable bounds on torque. (N-m)
problem.setControlInfo('/motorOA', [-200, 200]);

%% Configure Solver
solver = study.initCasADiSolver();
% solver = study.initTropterSolver();
solver.set_num_mesh_intervals(50);
solver.set_verbosity(2);
solver.set_optim_solver('ipopt');
solver.set_optim_convergence_tolerance(1e-4);
solver.set_optim_constraint_tolerance(1e-4);
solver.set_optim_max_iterations(1000);
guess = solver.createGuess();
guess.randomizeAdd();
solver.setGuess(guess);

%% Solve
sol = study.solve();

%% Save Motion
fn = [analysis_name dateLabel '.sto'];
sol.write([dirResults fn])

%% Visualize Motion
study.visualize(sol);

%% Plot Angles
close all;
S = sol.getStatesTrajectoryMat();
S_angles = S(:, 1:4);
Snames = sol.getStateNames();
t = sol.getTimeMat();

plot(t, rad2deg(S_angles))
legendLabels = cell([4 1]);
for i = 1:4
    legendLabels{i} = char(Snames.get(i - 1));
end
legend(legendLabels, 'Interpreter', 'none');

%% Plot Torque
U = sol.getControlsTrajectoryMat();
Unames = sol.getControlNames();

figure;
plot(t, U);
legend(char(Unames.get(0)));