%%% buildModel.m
%%%
%%% Builds the OSIM file for a simple four bar mechanism for use in testing
%%% kinematic loops in MoCo.
%%%
%%% Run this file to build the OSIM file. Then run FourBarP2P to optimize
%%% the trajectory of the point-to-point motion.
%%%
%%% Author:     Tyler Morrison
%%% Ohio State University - Dept. of Mechanical and Aerospace Engineering
%%% email:      morrison.730@osu.edu 
%%% Website:    http://www.tyler-morrison.com
%%% October 2020

%% Clear Workspace
clc; close all; clearvars;

%% Import OpenSim Libraries into Matlab
import org.opensim.modeling.*

%% Instantiate an (empty) OpenSim Model
modelName = 'FourBarTest';
model = Model();
model.setName(modelName);

% Get a reference to the ground object
ground = model.getGround();

% Define the acceleration of gravity
model.setGravity(Vec3(0, -9.81, 0));

%% Design parameters.
L_O = 1;  % Link O length. (Ground link).
L_A = 1;  % Link A length. (Motor input  link).
L_B = 1.1;% Link B length. (Coupler link).
L_C = 1;  % Link C length. (Output angle link).
width = 0.05; % Link width. (All links).
depth = 0.05; % Link depth. (All links).

% Diagram Not to Scale
%
%               Link B
%            o----------o
%           /            \
%   Link A /              \ Link C
%         /                \
%        o------------------o
%         xxxxxxxxxxxxxxxxxx
%            Link O/Ground

%% Build Links
% Link Colors: O = White, A = Red, B = Blue, C = Green 
linkO = buildLink('linkO', L_O, width, depth, 1, model, Vec3(1, 1, 1));
linkA = buildLink('linkA', L_A, width, depth, 1, model, Vec3(1, 0, 0));
linkB = buildLink('linkB', L_B, width, depth, 1, model, Vec3(0, 1, 0));
linkC = buildLink('linkC', L_C, width, depth, 1, model, Vec3(0, 0, 1));

%% Build Joints
%%% NOTE:
%%% This is the tricky bit that I messed up the first time. Since the four
%%% bar link age is closed chain mechanism (the links form a loop that
%%% touches the grounded link resulting in it only having one degree of
%%% freedom rather than 3) it requires careful treatment to run properly in
%%% OpenSim. In this case, there are two grounded objects: the ground link
%%% and the LinkO. The ground is connected to Link A with Joint OA while
%%% LinkO is connected to LinkC with Joint CO. Finally, the loop is closed
%%% by using the WeldConstraint to fuse linkO and ground. There may be more
%%% clever ways of doing it, but this worked.
%%%
%%% Suggestion from Nicholas Bianco set me straight.
%%%
%%% See forum post: 
%%% https://simtk.org/plugins/phpBB/viewtopicPhpbb.php?f=1815&t=12627

% Linkage Joints
[jointOA, OA_rz] = buildPinJoint(ground, linkA,...
    L_O, L_A, 'jointOA', 'OA_rz', model);
[jointAB, AB_rz] = buildPinJoint(linkA, linkB,...
    L_A, L_B, 'jointAB', 'AB_rz', model);
[jointBC, BC_rz] = buildPinJoint(linkB, linkC,...
    L_B, L_C, 'jointBC', 'BC_rz', model);
[jointCO, CO_rz] = buildPinJoint(linkC, linkO,...
    L_C, L_O, 'jointCO', 'CO_rz', model);


% WeldConstraint linkO to ground to close the kinematic chain.
locationInParent    = Vec3(0,0,0); orientationInParent = Vec3(0,0,0);
locationInChild     = Vec3(0,0,0); orientationInChild  = Vec3(0,0,0);
constraintOG = WeldConstraint('jointOG', linkO, locationInParent,...
    orientationInParent, ground, locationInChild, orientationInChild);
model.addComponent(constraintOG);

%% Build the Actuator
% Pure torque actuator between ground and Link A
bodyA = ground; bodyB = linkA;
axis = Vec3(0, 0, 1); axisInGround = false;
motorOA = TorqueActuator(bodyA, bodyB, axis, axisInGround);
motorOA.setName('motorOA');
model.addComponent(motorOA);

%% Initialize the System (checks model consistency).
model.initSystem();

%% Save the model to a file.
model.print([modelName '.osim']);

%% Local Functions
function link = buildLink(name, L, W, D, m, model, color)
%%% Simple function for building a link with length L, width W, depth D,
%%% and mass m. RGB Color is specified as Vec3.

import org.opensim.modeling.*

link = Body();
link.setName(name);
link.setMass(m);
link.setInertia( Inertia(1,1,1,0,0,0) );

linkGeo = Brick(Vec3(L/2, W/2, D/2));
linkGeo.setColor(color);

link.attachGeometry(linkGeo);

model.addComponent(link);

end

function [joint, coord] = buildPinJoint(parentBody, childBody,...
    L_parent, L_child, joint_name, coord_name, model)
%%% Simple function for building a Pin Joint between the parent and child
%%% links with lengths L_parent and L_child. The joint_name and coord_name
%%% should be specified.

import org.opensim.modeling.*

% Make and add a Pin joint for the parent to child
locationInParent    = Vec3(-L_parent/2,0,0);
orientationInParent = Vec3(0,0,0);
locationInChild     = Vec3(L_child/2,0,0);
orientationInChild  = Vec3(0,0,0);

joint = PinJoint(joint_name, parentBody, locationInParent,...
    orientationInParent, childBody, locationInChild, orientationInChild);


% Update the coordinates of the new joint
coord = joint.updCoordinate();
coord.setRange([deg2rad(-360), deg2rad(360)]);
coord.setName(coord_name);
coord.setDefaultValue(0);
coord.setDefaultSpeedValue(0);

% Add Joint to the Model
model.addComponent(joint);

end